
import ldap
from DSGlobals import *
from ldap.filter import escape_filter_chars, filter_format
import ldap.modlist
from copy import deepcopy
import string
import random
import smbpassword
import re
try:
    from crypt import crypt
except:
    from fcrypt import crypt

# NodeTypeTests should hold functions which decide whether to handle
# a particular ldap node, and return a DSNode if so, None if not. See
# further down for examples.
NodeTypeTests = []

def NodeFactory(parent, ldapData):
    """Create a DSNode object appropriate for the given ldapData.

    ldapData should be an object like that returned from an ldapobject.search.
    """
    global NodeTypeTests

    dn = ldapData[0]
    objectClasses = ldapData[1]['objectClass']
    tests = list(NodeTypeTests)
    tests.reverse()
    for test in tests:
        cls = test(parent, dn, objectClasses)
        if (cls):
            return cls(parent, ldapData)

re_whitespace = re.compile('\s')
def remove_whitespace(s):
    return re_whitespace.sub('', s)

class DSNode(object):
    """The base type for the various supported node types.
    
    The typical use of this object is as a dictionary with:
    node.GetInfo() # Initializes the values
    node['description'] = "interesting description"
    node.SetInfo() # Finalizes the changes
    """

    _interfaces = ("Node",)
    _attributes = ("Description",)
    
    def __init__(self, parent, ldapData):

        self.parent = parent
        self.conn = self.parent.conn
        self._dn = ldapData[0]
        self._objectClasses = ldapData[1]['objectClass']
        self.displayName = ldapData[1].get('displayName', [None])[0]
        self._cached_attributes = None
        self._new = None

        if not (hasattr(self, "_interfaces")):
            self._interfaces = ['Node']

    def __repr__(self):
        name = ldap.explode_dn(self._dn, 1)[0]
        return '<DSNode: %s>' % (name)

    def __eq__(self, other):
        if self._dn == other._dn:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, key):
        if not (key in self._attributes):
            raise KeyError(key)
        elif not (hasattr(self, "get_" + key)):
            raise KeyError(key)

        func = getattr(self, "get_" + key)
        return func()

    def __setitem__(self, key, value):
        if not (key in self._attributes):
            raise KeyError(key)

        elif not (hasattr(self, "set_" + key)):
            raise KeyError(key)

        func = getattr(self, "set_" + key)
        return func(value)

    def __contains__(self, key):
        return self._attributes.__contains__(key)

    def keys(self):
        return self._attributes

    def has_key(self, key):
        return key in self._attributes

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError, err:
            return default

    def _single_get(self, key):
        """Private method for getting single values"""
        if not (self._new):
            self.GetInfo()

        val = self._new.get(key, "")
        if not (type(val) == str):
            return val[0]
        else:
            return val

    def _single_set(self, key, val):
        """Private method for setting single values"""
        assert ((type(val) == str) or (type(val) == unicode))
        if not (self._new):
            self.GetInfo()
        if (val):
            self._new[key] = (val,)
        else:
            self._new[key] = val

    def _multi_get(self, key):
        """Private method for getting multi-values"""
        if not (self._new):
            self.GetInfo()

        val = self._new.setdefault(key, [])
        return deepcopy(val)

    def _multi_set(self, key, val):
        """Private method for setting multi-values"""
        assert ((type(val) == list) or (type(val) == tuple))
        if not (self._new):
            self.GetInfo()

        self._new[key] = val

    def _get_all(self):
        """Don't use this function unless you understand why you shouldn't use this function."""
        if not (self._new):
            self.GetInfo()
        return deepcopy(self._new)

    def IsInitialized(self):
        return self._new is not None

    def Implements(self, interface=None):
        """Return the list of interfaces implemented by this object. If an interface is specified, return True if it is implemented, otherwise False"""
        if (interface):
            return interface in self._interfaces
        return self._interfaces

    def Name(self):
        """Return the display name of the object if available. Otherwise return the first part of the dn"""
        name = ldap.explode_dn(self._dn, 1)[0]

        if (self._cached_attributes):
            name = self._cached_attributes.get('displayName', (name,))[0]
        elif (self.displayName):
            return self.displayName
        return name

    def Delete(self):
        self.conn.delete_s(self.Path())
        self._cached_attributes = None
        self._new = None
        self.parent.Refresh()

    def get_Description(self):
        return self._single_get("description")

    def set_Description(self, value):
        self._single_set("description", value)

    def Path(self):
        """Return the ldap path (dn) of the object"""
        return self._dn

    def PrettyPath(self):
        """Return a path using a filesystem-like representation. e.g. com/okmaybe/mrroach"""
        path = ldap.explode_dn(self._dn, 1)
        path.reverse()
        return '/'.join(path)

    def _apply_changes(self, modlist):
        try:
            resid = self.conn.modify(self._dn, modlist)
            rescode, results = self.conn.result(resid)
        except ldap.INSUFFICIENT_ACCESS, err:
            raise InsufficientAccessError, "Insufficient rights to perform this operation."
        except ldap.CONSTRAINT_VIOLATION, err:
            print err
            raise DSError, "The server reported a constraint violation."
            

    def GetInfo(self):
        """Initialize the object and undo any modifications that may have been made."""
        
        resid = self.conn.search(self._dn, ldap.SCOPE_BASE)
        rescode, results = self.conn.result(resid)
           
        # return the dict object(index of 1) from the first result(index of 0)
        self._cached_attributes = results[0][1]
        self._new = deepcopy(self._cached_attributes)
       
    def SetInfo(self):
        """Write the changes that have been made (if any) to the directory server."""
        if not (self._cached_attributes):
            return

        modlist = ldap.modlist.modifyModlist(self._cached_attributes, self._new)
        if not modlist:
            return

        self._apply_changes(modlist)
        self._cached_attributes = deepcopy(self._new)

def DSNodeTest(parent, dn, objectClasses):
    return DSNode
NodeTypeTests.append(DSNodeTest)
               
## objectClass implementations

class IDSGroup(DSNode):

    _interfaces = DSNode._interfaces + ('IDSGroup',)

    _attributes = DSNode._attributes + \
    (
        'GidNumber',
        'Members',
    )

    def get_GidNumber(self):
        return self._single_get("gidNumber")

    def set_GidNumber(self, value):
        self._single_set("gidNumber", value)

    def Add(self, userNode):
        """Add the specified userNode to the group"""
        uid = userNode._single_get("uid")
        memberUids = self._multi_get("memberUid")
        if not (uid in memberUids):
            memberUids.append(uid)
           
        self._multi_set("memberUid", memberUids)

    def Remove(self, userNode):
        """Remove the specified userNode from the group"""
        uid = userNode._single_get("uid")
        members = self._multi_get("memberUid")
        if (uid in members):
            del members[members.index(uid)]
        self._multi_set("memberUid", members)

    def Members(self, invert=False):
        """Return a list of nodes representing the members of this group, or if invert is True, the non-members of this group"""
        memberUids = self._multi_get("memberUid")

        if not (invert):
            filterTemplate = "(uid=%s)" * len(memberUids)
            primaryGroupFilter = "(&(objectClass=posixAccount)(gidNumber=%s))"
            searchstring = filter_format("(|" + filterTemplate + primaryGroupFilter + ")", \
                memberUids + [self._single_get('gidNumber')])
        else:
            filterTemplate = "(!(uid=%s))" * len(memberUids)
            primaryGroupFilter = "(!(gidNumber=%s))"
            searchstring = filter_format("(&(objectClass=posixAccount)" + filterTemplate + primaryGroupFilter + ")", \
                memberUids + [self._single_get('gidNumber')])
        
        
        nodes = self.parent.dsconn._search(searchstring)
        return nodes

    def set_Members(self):
        raise KeyError("Members is read-only. Use Add or Remove instead")

    def Delete(self):
        members = self.Members()
        for member in members:
            if member['PrimaryGroup'] == self:
                raise DSError, "User: %s has this group set as the primary group. Can't delete." % (member.PrettyPath())

        DSNode.Delete(self)

def IDSGroupTest(parent, dn, objectClasses):
    if 'posixGroup' in objectClasses:
        return IDSGroup
NodeTypeTests.append(IDSGroupTest)

class IDSUserList(DSNode):
    """Represents the groupOfNames objectClass"""
    _interfaces = DSNode._interfaces + ('IDSUserList',)
    _attributes = DSNode._attributes + \
    (
        'businessCategory',
        'description',
        'owner',
        'seeAlso'
    )

    def Add(self, userNode):
        """Add the specified userNode to the list."""
        members = self._multi_get("member")
        if not (userNode.Path() in members):
            members.append(userNode.Path())
            self._multi_set("member", members)

    def Remove(self, userNode):
        """Remove the specified userNode from the list"""
        members = self._multi_get("member")
        if (userNode.Path() in members):
            if (len(members) == 1):
                raise DSError, "User lists must contain at least one user."
            else:
                del members[members.index(userNode.Path())]
                self._multi_set("member", members)
    
    def Members(self, invert=False):
        """Return a list of nodes representing the members of this group, or if invert is True, the non-members of this group"""

        if not (invert):
            members = self._multi_get("member")
            nodes = self.parent.dsconn.GetObjects(members)
            # Make sure only users get passed along
            nodes = [n for n in nodes if n.Implements('IDSUser')]
        else:
            members = self._multi_get("member")
            users = self.parent.dsconn.Users()
            nodes = [n for n in users if n.Path() not in members]
        
        return nodes

        
def IDSUserListTest(parent, dn, objectClasses):
    if 'groupOfNames' in objectClasses:
        return IDSUserList
NodeTypeTests.append(IDSUserListTest)


class IDSUser(DSNode):

    _interfaces = DSNode._interfaces + ('IDSUser',)
    _attributes = DSNode._attributes + \
    (
        'Uid',
        'UidNumber',
        'Office',
        'FirstName',
        'LastName',
        'Initials',
        'TelephoneNumber',
        'EmailAddress',
        'DisplayName',
        'PrimaryGroup',
        'LoginShell',
        'HomeDirectory',
        'SambaLogonScript',
        'SambaHomePath',
        'SambaHomeDrive',
    )

    def get_Uid(self):
        return self._single_get("uid")

    def set_Uid(self, value):
        self._single_set("uid", value)

    def get_UidNumber(self):
        return self._single_get("uidNumber")

    def set_UidNumber(self, value):
        self._single_set("uidNumber", value) 

    def get_Office(self):
        return self._single_get("physicalDeliveryOfficeName")

    def set_Office(self, value):
        self._single_set("physicalDeliveryOfficeName", value)
    
    def get_FirstName(self):
        return self._single_get("givenName")

    def set_FirstName(self, value):
        self._single_set("givenName", value)

    def get_LastName(self):
        return self._single_get("sn")

    def set_LastName(self, value):
        self._single_set("sn", value)

    def get_Initials(self):
        return self._single_get("initials")

    def set_Initials(self, value):
        self._single_set("initials", value)

    def get_TelephoneNumber(self):
        return self._single_get("telephoneNumber")

    def set_TelephoneNumber(self, value):
        self._single_set("telephoneNumber", value)

    def get_EmailAddress(self):
        return self._single_get("mail")
        
    def set_EmailAddress(self, value):
        self._single_set("mail", value)

    def get_DisplayName(self):
        return self._single_get("displayName")

    def set_DisplayName(self, value):
        self._single_set("displayName", value)

    def get_LoginShell(self):
        return self._single_get("loginShell")

    def set_LoginShell(self, value):
        self._single_set("loginShell", value)

    def get_HomeDirectory(self):
        return self._single_get("homeDirectory")

    def set_HomeDirectory(self, value):
        self._single_set("homeDirectory", value)

    def get_SambaLogonScript(self):
        return self._single_get("sambaLogonScript")

    def set_SambaLogonScript(self, value):
        self._single_set("sambaLogonScript", value)

    def get_SambaHomeDrive(self):
        return self._single_get("sambaHomeDrive")

    def set_SambaHomeDrive(self, value):
        self._single_set("sambaHomeDrive", value)

    def get_SambaHomePath(self):
        return self._single_get("sambaHomePath")

    def set_SambaHomePath(self, value):
        self._single_set("sambaHomePath", value)

    def get_PrimaryGroup(self):
        gidNumber = self._single_get("gidNumber")

        searchstring = filter_format ("(&(objectClass=posixGroup)(gidNumber=%s))", (gidNumber,))
        
        nodes = self.parent.dsconn._search(searchstring)

        assert len(nodes) < 2, "More than one group with id %s exists" % (gidNumber)
        if (len(nodes) == 1):
            return nodes[0]

    def set_PrimaryGroup(self, groupNode):
        # This part is outside the whole SetInfo transaction, but
        # it seems safe to make someone a member of their primary
        # group... If we don't, they could stop being a member of
        # the group at all.
        oldPrimary = self['PrimaryGroup']
        if (oldPrimary and not self in oldPrimary.Members()):
            oldPrimary.Add(self)
            oldPrimary.SetInfo()
        self._single_set("gidNumber", groupNode._single_get("gidNumber"))

    def SetPassword(self, password):
        """Set the userNode password to the string specified by password.

        If the dsconnection is managing a kadmin server, the userPassword will
        be set to {SASL}uid@REALM and the kerberos password will be set.
        Otherwise, userPassword will be set using {crypt}. In both cases, the
        Samba password will be set.
        """
        attributes = self._get_all()
        assert ((type(password) == str) or (type(password) == unicode)), "SetPassword: password must be a string or unicode value"


        if (self.parent.dsconn.kadmin):
            # password changing should be done through kerberos
            kadmin = self.parent.dsconn.kadmin
            uid = self._single_get('uid')
            princ = '%s@%s' % (uid, kadmin.realm)
            try:
                kadmin.GetPrincipal(princ)
            except:
                try:
                    kadmin.CreatePrincipal(princ, password)
                except:
                    raise DSError, "No such principal: %s" % (princ)
            else:
                try:
                    kadmin.SetPassword(princ, password)
                except:
                    raise DSError, "Unable to set password"
            attributes['userPassword'] = ('{SASL}%s' % (princ),)
        else:
            salt_chars = string.ascii_letters + string.digits + '/.'
            salt = salt_chars[random.randint(0, len(salt_chars) - 1)] + \
                salt_chars[random.randint(0, len(salt_chars) - 1)]
            cryptpw = '{crypt}' + crypt(password, salt)

            attributes['userPassword'] = (cryptpw,)
        if ('sambaSAMAccount' in self._objectClasses):
            ntpassword = smbpassword.nthash(password)
            lmpassword = smbpassword.lmhash(password)

            attributes["sambaNTPassword"] = (ntpassword,)
            attributes["sambaLMPassword"] = (lmpassword,)
        
        modlist = ldap.modlist.modifyModlist(self._cached_attributes, attributes)
        if not modlist:
            return

        self._apply_changes(modlist)
        if ('sambaSAMAccount' in self._objectClasses):
            # I am cheating here to avoid having to run SetInfo and apply
            # all the unrelated changes that may have been made
            self._cached_attributes['sambaNTPassword'] = (ntpassword,)
            self._new['sambaNTPassword'] = (ntpassword,)
            self._cached_attributes['sambaLMPassword'] = (lmpassword,)
            self._new['sambaLMPassword'] = (lmpassword,)

    def Delete(self):

        if (self.parent.dsconn.kadmin):
            # kerberos principal should be deleted
            kadmin = self.parent.dsconn.kadmin
            uid = self._single_get('uid')
            princ = '%s@%s' % (uid, kadmin.realm)
            try:
                kadmin.GetPrincipal(princ)
            except:
                print "Warning: No such principal; %s" % (princ)
            else:
                try:
                    kadmin.DeletePrincipal(princ)
                except:
                    raise DSError, "Unable to delete kerberos principal: %s" % (princ)
        groups = self.Groups()
        for group in groups:
            group.Remove(self)
            group.SetInfo()

        DSNode.Delete(self)


    def Groups(self, invert=False):
        """Return the group Nodes for the user. Use invert=True for the groups that don't include the user"""
        
        gidNumber = self._single_get("gidNumber")
        uid = self._single_get("uid")

        if (invert):
            searchstring = filter_format(
                '''(|(&(objectClass=posixGroup)(!(gidNumber=%s))(!(memberUid=%s)))(&(objectClass=groupOfNames)(!(member=%s))))''',
                (gidNumber, uid, self._dn))

        else:
            searchstring = filter_format(
                '''(|(&(objectClass=posixGroup)(|(memberUid=%s)(gidNumber=%s)))(&(objectClass=groupOfNames)(member=%s)))''', (uid, gidNumber, self._dn))
        
        return self.parent.dsconn._search(searchstring)

def IDSUserTest(parent, dn, objectClasses):
    if ('posixAccount' in objectClasses):
        return IDSUser
NodeTypeTests.append(IDSUserTest)


class IDSComputer(DSNode):
    """Represents a samba computer account"""

    _interfaces = DSNode._interfaces + ('IDSComputer',)
    _attributes = DSNode._attributes + \
    (
        'UidNumber',
    )

    def get_UidNumber(self):
        return self._single_get("uidNumber")

    def set_UidNumber(self, value):
        self._single_set("uidNumber", value)

    def Name(self):
        # new name function that leaves off the $
        name = DSNode.Name(self)
        if name.endswith('$'):
            return name[:-1]
        return name
        
def IDSComputerTest(parent, dn, objectClasses):
    if ('posixAccount' in objectClasses):
        name = ldap.explode_dn(dn, 1)[0]
        if (name.endswith('$')):
            return IDSComputer
NodeTypeTests.append(IDSComputerTest)



class IDSUnixIDPool(DSNode):

    _interfaces = DSNode._interfaces + ('IDSUnixIDPool',)

    _attributes = DSNode._attributes + \
    (
        'GidNumber',
        'UidNumber',
    )
    
    ldapSearchFilter = "(objectClass=sambaUnixIDPool)"

    def get_GidNumber(self):
        return self._single_get("gidNumber")

    def set_GidNumber(self, value):
        self._single_set("gidNumber", value)

    def get_UidNumber(self):
        return self._single_get("uidNumber")

    def set_UidNumber(self, value):
        self._single_set("uidNumber", value)


def IDSUnixIDPoolTest(parent, dn, objectClasses):
    if 'sambaUnixIdPool' in objectClasses:
        return IDSUnixIDPool
NodeTypeTests.append(IDSUnixIDPoolTest)

class IDSSambaDomain(DSNode):

    _interfaces = DSNode._interfaces + ('IDSSambaDomain',)

    _attributes = DSNode._attributes + \
    (
        'SambaSID',
    )
    
    ldapSearchFilter = "(objectClass=sambaDomain)"

    def get_SambaSID(self):
        return self._single_get("sambaSID")

    def set_SambaSID(self, value):
        self._single_set("sambaSID", value)

def IDSSambaDomainTest(parent, dn, objectClasses):
    if 'sambaDomain' in objectClasses:
        return IDSSambaDomain
NodeTypeTests.append(IDSSambaDomainTest)

